#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['crabnet',
 'crabnet.data',
 'crabnet.data.benchmark_data',
 'crabnet.data.benchmark_data.CritExam__Ed',
 'crabnet.data.benchmark_data.CritExam__Ef',
 'crabnet.data.benchmark_data.OQMD_Bandgap',
 'crabnet.data.benchmark_data.OQMD_Energy_per_atom',
 'crabnet.data.benchmark_data.OQMD_Formation_Enthalpy',
 'crabnet.data.benchmark_data.OQMD_Volume_per_atom',
 'crabnet.data.benchmark_data.aflow__Egap',
 'crabnet.data.benchmark_data.aflow__ael_bulk_modulus_vrh',
 'crabnet.data.benchmark_data.aflow__ael_debye_temperature',
 'crabnet.data.benchmark_data.aflow__ael_shear_modulus_vrh',
 'crabnet.data.benchmark_data.aflow__agl_thermal_conductivity_300K',
 'crabnet.data.benchmark_data.aflow__agl_thermal_expansion_300K',
 'crabnet.data.benchmark_data.aflow__energy_atom',
 'crabnet.data.benchmark_data.mp_bulk_modulus',
 'crabnet.data.benchmark_data.mp_e_hull',
 'crabnet.data.benchmark_data.mp_elastic_anisotropy',
 'crabnet.data.benchmark_data.mp_mu_b',
 'crabnet.data.benchmark_data.mp_shear_modulus',
 'crabnet.data.element_properties',
 'crabnet.data.matbench',
 'crabnet.data.materials_data',
 'crabnet.data.materials_data.elasticity',
 'crabnet.data.materials_data.example_materials_property',
 'crabnet.utils']

package_data = \
{'': ['*'],
 'crabnet': ['tests/*'],
 'crabnet.data': ['matbench_cv/*',
                  'matbench_cv/castelli/*',
                  'matbench_cv/dielectric/*',
                  'matbench_cv/elasticity_log10(G_VRH)/*',
                  'matbench_cv/elasticity_log10(K_VRH)/*',
                  'matbench_cv/expt_gap/*',
                  'matbench_cv/expt_is_metal/*',
                  'matbench_cv/glass/*',
                  'matbench_cv/jdft2d/*',
                  'matbench_cv/mp_e_form/*',
                  'matbench_cv/mp_gap/*',
                  'matbench_cv/mp_is_metal/*',
                  'matbench_cv/phonons/*',
                  'matbench_cv/steels_yield/*']}

install_requires = \
['numpy',
 'pandas',
 'scikit-learn',
 'scipy',
 'matplotlib',
 'seaborn',
 'tqdm',
 'joblib',
 'psutil']

setup(name='crabnet',
      version='1.0.6',
      description='Predict materials properties using only the composition information.',
      author=None,
      author_email='"Sterling G. Baird" <sterling.baird@utah.edu>, Anthony Yu-Tung Wang <aytwang@alumni.uwaterloo.ca>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.7,<3.10',
     )
